
#include <string.h>
#include <stdarg.h>

#include "ncommon.h"
#include "nstring.h"
#include "shell.h"
#include "console.h"
#include "cvar.h"

#include "shell_local.h"


ShellState Shell;

Shell_Function * Shell_Find(const char *name) {
	Hash_Data tmp;
	tmp.name = name;
	tmp.name_len = String_Length( tmp.name );
	return (Shell_Function *) Hash_Get( Shell.functionHash, &tmp );
}

static void Shell_Hash_Do_Free(UNUSED void *hash, Shell_Function *function) {
	String_Free( (char **) (char *) &function->h.name );
	String_Free( (char **) (char *) &function->description );
	Mem_Free(  &function );
}

void Shell_RegisterFunction( const char *name, Shell_Function_CB callback, const char *description ) {
	Shell_Function *function;

	if( Shell_FunctionExists( name ) ) {
		Console_Warning( "Shell functions '%s' exists already!\n", name );
		return;
	}

	// add the new function
	function = Mem_Alloc( Shell.zone, sizeof( Shell_Function ) );
	function->h.name_len = String_Set( (char **) (char *) &function->h.name, Shell.zone, name );
	function->callback = callback;
	String_Set( (char **) (char *)&function->description, Shell.zone, description );

	Hash_Add( Shell.functionHash, &function->h );
}

void Shell_AddText( const char *text ) {
	String_Append( &Shell.commandText, Shell.zone, text);
}

Hash_Sort_Item *Shell_GetList( void ) {
	return Hash_CreateSortedList( Global_Clump, Shell.functionHash );
}

void Shell_Init(void) {
	if( Shell.active ) {
		Console_Warning( "Shell has already been initialized!\n" );
		return;
	}

	memset( &Shell, 0, sizeof(Shell) );
	Shell.active = true;
	Shell.zone = Mem_AllocZone( "Shell", NULL, 0 );
	Shell.interpreterZone = Mem_AllocClump( "Interpreter", Shell.zone, (1 << 15), 0 );
	Shell.functionHash = Hash_Create( 10, Shell.zone, NULL, NULL, (Do_Free *) Shell_Hash_Do_Free );
}

void Shell_Quit(void) {
	if( !Shell.active ) {
		Console_Warning( "Tried to quit uninitialized Shell system!\n" );
		return;
	}

	Hash_Destroy( Shell.functionHash );
	Mem_FreeZone( &Shell.zone );
	memset( &Shell, 0, sizeof(Shell) );
	Shell.active = false;
}

// TODO: merge this parser with the Util_ParseC parser
#define SHELL_PARSER_SPECIAL_CHARS "{}();\""

static inline Nbool IsNewLine( const char c ) {
	return c == '\r' || c == '\n';
}

static inline Nbool IsWhitespace( const char c ) {
	return c && c <= ' ';
}

// FIXME: there has to be a clib function for this
static inline Nbool IsInCharSet( const char c, const char *charSet ) {
	for( ; *charSet ; charSet++ ) {
		if( c == *charSet ) {
			return true;
		}
	}
	return false;
}

static inline Nbool Shell_Parser_IsSimpleToken( const char *text ) {
	const char *current;
	for( current = text ; *current ; current++ ) {
		if( *current <= ' ' || IsInCharSet( *current, SHELL_PARSER_SPECIAL_CHARS ) ) {
			return false;
		}
	}
	return true;
}

static inline Nbool Shell_Parser_MatchText( Shell_Parser *parser, const char *text ) {
	NUint length = strlen( text );

	if( strncmp( parser->currentChar, text, length ) == 0 ) {
		parser->currentChar += length;
		return true;
	}
	return false;
}

/*
static inline Nbool Shell_Parser_MatchPair( Shell_Parser *parser, const char first, const char second] ) {
	if( parser->currentChar[0] == first && parser->currentChar[1] == second ) {
		parser->currentChar += 2;
		return true;
	}
	return false;
}
*/

static inline Nbool Shell_Parser_MatchChar( Shell_Parser *parser, const char c ) {
	if( *parser->currentChar == c ) {
		parser->currentChar++;
		return true;
	}
	return false;
}

static inline void Shell_Parser_ReadNewLines( Shell_Parser *parser ) {
	while( 1 ) {
		if( Shell_Parser_MatchChar( parser, '\n' ) ) {
			parser->line++;
		} else if( Shell_Parser_MatchText( parser, "\r\n" ) ) {
			parser->line++;
		} else {
			break;
		}
	}
}

static inline void Shell_Parser_ReadNextWithNewLines( Shell_Parser *parser ) {
	if( IsNewLine( *parser->currentChar ) ) {
		Shell_Parser_ReadNewLines( parser );
	}
	if( *parser->currentChar ) { 
		*parser->currentChar++;
	}
}

// skips whitespace and // comments and /* */ comments
// result: you'll have a meaningfull char in currentChar
static void Shell_Parser_SkipStuff( Shell_Parser *parser ) {
	while( 1 ) {
		while( IsWhitespace( *parser->currentChar ) ) {
			Shell_Parser_ReadNextWithNewLines( parser );
		}

		if( Shell_Parser_MatchText( parser, "//" ) ) {
			while( *parser->currentChar && !IsNewLine( *parser->currentChar ) ) {
				parser->currentChar++;
			}
		} else if( Shell_Parser_MatchText( parser, "/*" ) ) {
			while( *parser->currentChar && !Shell_Parser_MatchText( parser, "*/" ) ) {
				Shell_Parser_ReadNextWithNewLines( parser );
			}
		} else { 
			break;
		}
	}
}

static inline void Shell_Parser_ReadString( Shell_Parser *parser ) {
	while( *parser->currentChar && !Shell_Parser_MatchChar( parser, '"' ) ) {
		// hrm, this isnt supposed to be clever 
		Shell_Parser_MatchChar( parser, '\\' );
		Shell_Parser_ReadNextWithNewLines( parser );
	}
}

// whitespace always stops the parsing 
static inline void Shell_Parser_ReadToken( Shell_Parser *parser, const char *stopCharSet ) {
	while( *parser->currentChar && *parser->currentChar > ' ' && 
		!IsInCharSet( *parser->currentChar, stopCharSet ) ) {
		parser->currentChar++;
	}		
}

// reads through a whole block with correct string handling
// e.g. pass {, } and it will correctly parse:
// { "{}}" { a ( 1.0 c ) } "}{" } as one block
static void Shell_Parser_ReadBlock( Shell_Parser *parser, const char startChar, const char endChar ) {
	NUint depth = 1;
	while( depth > 0 && *parser->currentChar ) {
		Shell_Parser_SkipStuff( parser );
		if( Shell_Parser_MatchChar( parser, '"' ) ) {
			Shell_Parser_ReadString( parser );
		} else if( Shell_Parser_MatchChar( parser, startChar ) ) {
			depth++;
		} else if( Shell_Parser_MatchChar( parser, endChar ) ) {
			depth--;
		} else {
			// ignore the char
			Shell_Parser_ReadNextWithNewLines( parser );
		}
	}
}

static void Shell_Parser_SaveToken( Shell_Parser *parser, const char *start, const char *afterEnd, Nbool replaceEscapeSequences ) {
	if( !replaceEscapeSequences ) {
		String_Substring( &parser->token, Shell.zone, start, 0, afterEnd - start );
	} else {
		const char *in;
		char *out;

		// an escape sequence actually removes at least one char from the buffer (the first \)
		// so we can just allocate the buffer now and then copy it in one by one
		out = parser->token = String_Alloc( Shell.zone, afterEnd - start );

		for( in = start ; in != afterEnd ; in++, out++ ) {
			// this is ok because of the trailing \0 
			if( in[0] != '\\' && IsInCharSet( in[1], "rnt" ) ) {
				in++;
				switch( *in ) {
				case 'r':
					*out = '\r';
					break;
				case 'n':
					*out = '\n';
					break;
				case 't':
					*out = '\t';
					break;
				}
			} else {
				*out = *in;
			}
		}
	}
}

/*
Reads in { } blocks as one string (and treats it as block, see ReadBlock)
Reads in "" correctly and parses some escape sequences
Reads in (, ) and ; as single tokens
Reads normal text as token (whitespace is a delimiter)
*/
Nbool Shell_Parser_Parse( Shell_Parser *parser ) {
	const char *start, *afterEnd;
	Nbool replaceEscapeSequences;

	String_Free( &parser->token );

	if( *parser->currentChar == '\0' ) {
		Console_Warning( "Already reached end of Shell buffer!\n" );
		return false;
	}

	Shell_Parser_SkipStuff( parser );

	start = parser->currentChar;
	replaceEscapeSequences = false;

	if( Shell_Parser_MatchChar( parser, '{' ) ) {
		start++;
		Shell_Parser_ReadBlock( parser, '{', '}' );
	} else if( Shell_Parser_MatchChar( parser, '"' ) ) {
		start++;
		Shell_Parser_ReadString( parser );
		replaceEscapeSequences = true;
	} else if( Shell_Parser_MatchChar( parser, '(' ) ||
		Shell_Parser_MatchChar( parser, ')' ) ||
		Shell_Parser_MatchChar( parser, ';' ) ) {
		// we're done here, see the end of the function
	} else {
		Shell_Parser_ReadToken( parser, SHELL_PARSER_SPECIAL_CHARS );
	}
	afterEnd = parser->currentChar;

	if( start == afterEnd ) {
		return false;
	}

	Shell_Parser_SaveToken( parser, start, afterEnd, replaceEscapeSequences );	
	return true;
}

static inline void Shell_Parser_Init( Shell_Parser *parser, const char *text, NUint lineBias ) {
	assert( text );

	parser->currentChar = text;
	parser->line = lineBias;
	parser->token = NULL;
}

static inline void Shell_Parser_Destory( Shell_Parser *parser ) {
	String_Free( &parser->token );
}

/*
<block> ::= "(" <statements> ")"
<statements> ::= <statement> (";" <statement> )* [";"]
<statement> ::= <item>*
<item> ::= <symbol> | <block> | <lateblock>
<lateblock> ::= "{" <statement>* "}"
*/

static inline void Shell_Statement_Init( Shell_Statement *statement ) {
	memset( statement, 0, sizeof( *statement ) );

	assert( Shell.currentBlock->currentStatement == NULL );
	Shell.currentBlock->currentStatement = statement;
}

static inline char * Shell_Statement_GetFirst( Shell_Statement *statement ) {
	// test listSize!
	return statement->list[ 0 ];
}

static inline char ** Shell_Statement_GetNew( Shell_Statement *statement ) {
	//FIXME: test for max list size!
	return &statement->list[ statement->listSize++ ];
}

static inline void Shell_Statement_Destroy( Shell_Statement *statement ) {
	NUint i;

	for( i = 0 ; i < statement->listSize ; i++ ) {
		String_Free( &statement->list[ i ] );
	}

	assert( Shell.currentBlock->currentStatement != NULL );
	Shell.currentBlock->currentStatement = NULL;
}

static inline void Shell_Block_Init( Shell_Block *block, Shell_Parser *currentParser ) {
	memset( block, 0, sizeof( Shell_Block ) );

	block->returnText = NULL;
	block->parser = currentParser;
	block->currentStatement = NULL;

	block->parent = Shell.currentBlock;
	Shell.currentBlock = block;
	Shell.executionDepth++;
}

static inline void Shell_Block_Destroy( Shell_Block *block ) {
	String_Free( &block->returnText );

	Shell.executionDepth--;
	Shell.currentBlock = block->parent;
}

void Shell_Exec_Error( const char *msg, ... ) {
	//TODO
	Console_Printf( "%i: Error!\n", Shell.currentBlock->parser->line );
}

static void Shell_Interpreter_InterpretStatement( ) {
	Shell_Parser *parser = Shell.currentBlock->parser;
	Shell_Statement statement;
	
	Shell_Statement_Init( &statement );
	
	while( *parser->token && 
		!(String_Length( parser->token ) == 1 && IsInCharSet( *parser->token, SHELL_PARSER_SPECIAL_CHARS )) ) {
		if( String_Compare( parser->token, "(" ) ) {
			Shell_Interpreter_InterpretSubBlock( Shell.interpreterZone, Shell_Statement_GetNew( &statement ) );
		} else {
			String_Set( Shell_Statement_GetNew( &statement ), Shell.interpreterZone, parser->token );
		}
		Shell_Parser_Parse( parser );
	}

	if( statement.listSize == 0 ) {
		// FIXME: every statement should set the frame's returnText
		return;
	}

	// TODO: error out if !*token
	if( !Shell_Parser_IsSimpleToken( Shell_Statement_GetFirst( &statement ) ) ) {
		Shell_Interpreter_InterpretTextBlock( Shell_Statement_GetFirst( &statement ), Shell.zone, &Shell.currentBlock->returnText );
	} else {
		// call shell function
	}
}

void Shell_Interpreter_InterpretStatements() {
	Shell_Parser *parser = Shell.currentBlock->parser;

	Shell_Parser_Parse( parser );

	// FIXME: check for !*token
	while( 1 ) {
		Shell_Interpreter_InterpretStatement();
		if( String_Compare( parser->token, ";" ) == 0 ) {
			Shell_Parser_Parse( parser );
		} else {
			break;
		}
	}
}

void Shell_Interpreter_InterpretBlock() {
	Shell_Parser *parser = Shell.currentBlock->parser;

	Shell_Interpreter_InterpretStatements();

	if( String_Compare( parser->token, ")" ) != 0 ) {
		// FIXME: error!
	}
}

void Shell_Interpreter_InterpretText() {
	Shell_Parser *parser = Shell.currentBlock->parser;

	Shell_Interpreter_InterpretStatements();

	if( parser->token != '\0' ) {
		// FIXME: error!
	}
}

void Shell_Interpreter_InterpretSubBlock( Mem_Zone *zone, char **returnOut ) {
	Shell_Block block;

	Shell_Block_Init( &block, Shell.currentBlock->parser );

	Shell_Interpreter_InterpretBlock();

	if( returnOut ) {
		String_Set( returnOut, zone, block.returnText );
	}

	Shell_Block_Destroy( &block );
}

void Shell_Interpreter_InterpretTextBlock( const char *text, Mem_Zone *zone, char **returnOut ) {
	Shell_Parser parser;
	Shell_Block block;
	
	assert( text );

	Shell_Parser_Init( &parser, text, 1 );
	Shell_Block_Init( &block, &parser );

	Shell_Interpreter_InterpretBlock();

	if( returnOut ) {
		String_Set( returnOut, zone, block.returnText );
	}

	Shell_Block_Destroy( &block );
	Shell_Parser_Destory( &parser );
}

void Shell_ExecuteText( void ) {
	if( Shell.commandText ) {
		Shell_Parser parser;
		Shell_Block block;

		assert( Shell.currentFrame == NULL );
		assert( Shell.depth == 0 );

		assert( text );

		Shell_Parser_Init( &parser, text, 1 );
		Shell_Block_Init( &block, &parser );

		Shell_Interpreter_InterpretBlock();

		if( returnOut ) {
			String_Set( returnOut, zone, block.returnText );
		}

		Shell_Block_Destroy( &block );
		Shell_Parser_Destory( &parser );			

		String_Free( &Shell.commandText );
	}
}


/*
void Command_Execute(const char *text, const char *textend)
{
	NUint i;
	char *temptoken;
	const char *s;
	Cmd *c;
	s = text;
	Command.argc = 0;
	temptoken = NULL;
	for (;;)
	{
		if (!Command_GetToken(&s, textend, &temptoken))
			break;
		if (Command.argc < MAX_COMMANDPARAMETERS)
		{
			Command.argv[Command.argc++] = temptoken;
			temptoken = NULL;
		}
	}
	String_Free(&temptoken);
	if (Command.argc > 0)
	{
		if ((c = Command_Find(Command.argv[0])))
			c->func();
		else if (!Cvar_Command()) // pass it to the cvar module perhaps it can do something
			Console_Printf("unknown command \"%s@0\"\n", Command.argv[0]);
	}
	for (i = 0;i < Command.argc;i++)
		String_Free(&Command.argv[i]);
}
*/

/*
Nbool Shell_GetToken( const char **textPointer, const char *textEnd, char **tokenPointer )
{
	const char *currentChar;

	Nbool quoted;
	NUint pos;

	char tokenbuf[1024];

	currentChar = *textpointer;
	// unless it's a valid token, assume it hit the end
	*textpointer = textend;
	String_Set(tokenpointer, Shell.interpreterZone, NULL);
	while( currentChar < textEnd ) {
		// skip leading whitespace
		while( *currentChar && *currentChar <= ' ' ) {
			currentChar++;
		}
		// skip // comments
		if(currentChar < textend - 2 && currentChar[0] == '/' && currentChar[1] == '/')
		{
			currentChar += 2;
			for (;;)
			{
				if (currentChar >= textend)
					return false;
				if (currentChar[0] == '\r' || currentChar[1] == '\n')
				{
					currentChar++;
					break;
				}
				currentChar++;
			}
			continue;
		}
		// skip /*  *\/  comments
		if (currentChar < textEnd - 4 && *currentChar == '/' && currentChar[1] == '*')
		{
			currentChar += 2;
			for (;;)
			{
				if (currentChar > textend - 2)
					return false;
				if (*currentChar == '*' && currentChar[1] == '/')
				{
					currentChar += 2;
					break;
				}
				s++;
			}
			continue;
		}
		// parse a string, quoted or not
		if (*currentChar == '"')
		{
			quoted = true;
			currentChar++;
		}
		else
			quoted = false;
		pos = 0;
		for (;;)
		{
			if (quoted)
			{
				// quoted strings end with a quote
				if (s >= textend)
					return false;
				if (*currentChar == '"')
				{
					currentChar++;
					break;
				}
			}
			else
			{
				// unquoted strings end with whitespace or newline
				if (currentChar >= textEnd || *currentChar <= ' ')
					break;
			}
			// insert a character
			if (*currentChar == '\\')
			{
				// escape code, check which one
				currentChar++;
				if (*currentChar == 'r')
					tokenbuf[pos++] = '\r';
				else if (*currentChar == 'n')
					tokenbuf[pos++] = '\n';
				else if (*currentChar == 't')
					tokenbuf[pos++] = '\t';
				else
					tokenbuf[pos++] = *currentChar;
			}
			else
				tokenbuf[pos++] = *currentChar;
			currentChar++;
			// if tokenbuf is full, append it to the token
			if (pos >= sizeof(tokenbuf) - 1)
			{
				tokenbuf[pos] = 0;
				String_Append(tokenPointer, Shell.interpreterZone, tokenbuf);
				pos = 0;
			}
		}
		// if tokenbuf is not empty, append it to the token
		if (pos > 0)
		{
			tokenbuf[pos] = 0;
			String_Append(tokenPointer, Shell.interpreterZonee, tokenbuf);
			pos = 0;
		}
		// parsed a token
		*textPointer = currentChar;
		return true;
	}
	// failed, hit end of text
	return false;
}
*/